# Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestMeta(test_case.DslTestCase):
    def setUp(self):
        super(TestMeta, self).setUp()
        self._runner = self.new_runner(om.Object('metatests.TestMeta'))

    def test_class_multi_meta(self):
        self.assertCountEqual(
            [4, 1, 111, 2], self._runner.testClassMultiMeta())

    def test_class_single_meta(self):
        self.assertCountEqual(
            [5, 6], self._runner.testClassSingleMeta())

    def test_parent_class_not_inherited_meta(self):
        self.assertEqual(3, self._runner.testParentClassNotInheritedMeta())

    def test_method_meta(self):
        self.assertCountEqual(
            [7, 8, 9, 4, 1, 10], self._runner.testMethodMeta())

    def test_method_argument_meta(self):
        self.assertCountEqual(
            [1, 2, 3], self._runner.testMethodArgumentMeta())

    def test_inherited_property_meta(self):
        self.assertEqual(
            [1], self._runner.testInheritedPropertyMeta())

    def test_overridden_property_meta(self):
        self.assertCountEqual(
            [1, 4], self._runner.testOverriddenPropertyMeta())

    def test_package_meta(self):
        self.assertEqual(
            [], self._runner.testPackageMeta())

    def test_complex_meta(self):
        self.assertCountEqual([
            [1, 'metatests.PropertyType'],
            [2, 'metatests.PropertyType'],
            [3, 'metatests.PropertyType2'],
            [4, 'metatests.PropertyType'],
            [5, 'metatests.PropertyType2']
        ], self._runner.testComplexMeta())



# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import socket

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging

from murano.common import rpc
from murano.common import uuidutils
from murano.dsl import dsl

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


@dsl.name('io.murano.system.StatusReporter')
class StatusReporter(object):
    def __init__(self, environment):
        if not rpc.initialized():
            rpc.init()
        self._notifier = messaging.Notifier(
            rpc.NOTIFICATION_TRANSPORT,
            publisher_id=uuidutils.generate_uuid(),
            topics=['murano'])
        if isinstance(environment, str):
            self._environment_id = environment
        else:
            self._environment_id = environment.id

    def _report(self, instance, msg, details=None, level='info'):
        body = {
            'id': (self._environment_id if instance is None
                   else instance.id),
            'text': msg,
            'details': details,
            'level': level,
            'environment_id': self._environment_id,
            'timestamp': datetime.utcnow().isoformat()
        }
        self._notifier.info({}, 'murano.report_notification', body)

    def report(self, instance, msg):
        self._report(instance, msg)

    def report_error_(self, instance, msg):
        self._report(instance, msg, None, 'error')

    @dsl.name('report_error')
    def report_error(self, instance, msg):
        self._report(instance, msg, None, 'error')


class Notification(object):
    def __init__(self):
        if not CONF.stats.env_audit_enabled:
            return
        if not rpc.initialized():
            rpc.init()
        self._notifier = messaging.Notifier(
            rpc.NOTIFICATION_TRANSPORT,
            publisher_id=('murano.%s' % socket.gethostname()),
            driver='messaging')

    def _report(self, event_type, environment, level='info'):
        if not CONF.stats.env_audit_enabled:
            return

        if 'deleted' in environment:
            deleted_at = environment['deleted'].isoformat()
        else:
            deleted_at = None

        body = {
            'id': environment['id'],
            'level': level,
            'environment_id': environment['id'],
            'tenant_id': environment['tenant_id'],
            'created_at': environment.get('created').isoformat(),
            'deleted_at': deleted_at,
            'launched_at': None,
            'timestamp': datetime.utcnow().isoformat()
        }

        optional_fields = ("deployment_started", "deployment_finished")
        for f in optional_fields:
            body[f] = environment.get(f, None)

        LOG.debug("Sending out notification, type=%s, body=%s, level=%s",
                  event_type, body, level)

        self._notifier.info({}, 'murano.%s' % event_type,
                            body)

    def report(self, event_type, environment):
        self._report(event_type, environment)

    def report_error(self, event_type, environment):
        self._report(event_type, environment, 'error')


NOTIFIER = None


def get_notifier():
    global NOTIFIER
    if not NOTIFIER:
        NOTIFIER = Notification()

    return NOTIFIER



# Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from testtools import matchers
from unittest import mock
from yaql.language import exceptions as yaql_exceptions

from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case

from castellan.common import exception as castellan_exception


class TestEngineYaqlFunctions(test_case.DslTestCase):
    def setUp(self):
        super(TestEngineYaqlFunctions, self).setUp()
        self._runner = self.new_runner(om.Object('TestEngineFunctions'))

    def test_join(self):
        self.assertEqual('xx 123', self._runner.testJoin())

    def test_split(self):
        self.assertEqual(
            ['x', 'yy', '123'],
            self._runner.testSplit())

    def test_len(self):
        self.assertEqual(8, self._runner.testLen())

    def test_coalesce(self):
        self.assertEqual('a', self._runner.testCoalesce('a', 'b', 'c'))
        self.assertEqual('b', self._runner.testCoalesce(None, 'b', 'c'))
        self.assertEqual('c', self._runner.testCoalesce(None, None, 'c'))

    def test_base64(self):
        encoded = 'VEVTVA=='
        self.assertEqual(
            encoded,
            self._runner.testBase64Encode('TEST'))
        self.assertEqual(
            'TEST',
            self._runner.testBase64Decode(encoded))

    def test_format(self):
        self.assertEqual(
            '2 + 3',
            self._runner.testFormat('{0} + {1}', 2, 3))

    def test_replace_str(self):
        self.assertEqual(
            'John Doe',
            self._runner.testReplaceStr('John Kennedy', 'Kennedy', 'Doe'))

        self.assertRaises(
            yaql_exceptions.NoMatchingMethodException,
            self._runner.testReplaceStr, None, 'Kennedy', 'Doe')

    def test_replace_dict(self):
        self.assertEqual(
            'Marilyn Monroe',
            self._runner.testReplaceDict('John Kennedy', {
                'John': 'Marilyn',
                'Kennedy': 'Monroe'
            }))

    def test_lower(self):
        self.assertEqual(
            'test',
            self._runner.testToLower('TESt'))

    def test_upper(self):
        self.assertEqual(
            'TEST',
            self._runner.testToUpper('tEst'))

    def test_starts_with(self):
        self.assertIs(
            True,
            self._runner.testStartsWith('TEST', 'TE'))
        self.assertIs(
            False,
            self._runner.testStartsWith('TEST', 'te'))

    def test_ends_with(self):
        self.assertIs(
            True,
            self._runner.testEndsWith('TEST', 'ST'))
        self.assertIs(
            False,
            self._runner.testEndsWith('TEST', 'st'))

    def test_trim(self):
        self.assertEqual(
            'test',
            self._runner.testTrim('\n\t test \t\n'))

    def test_substr(self):
        self.assertEqual(
            'teststr',
            self._runner.testSubstr('teststr', 2, 3))

    def test_str(self):
        self.assertEqual(
            '123',
            self._runner.testStr(123))
        self.assertEqual(
            'true',
            self._runner.testStr(True))
        self.assertEqual(
            'false',
            self._runner.testStr(False))
        self.assertEqual(
            'null',
            self._runner.testStr(None))

    def test_int(self):
        self.assertEqual(
            123,
            self._runner.testInt('123'))
        self.assertEqual(
            0,
            self._runner.testInt(None))

    def test_keys(self):
        self.assertThat(
            self._runner.testKeys({True: 123, 5: 'Q', 'y': False}),
            matchers.MatchesSetwise(
                matchers.Equals('y'),
                matchers.Is(True),
                matchers.Equals(5)))

    def test_values(self):
        self.assertThat(
            self._runner.testValues({True: 123, 5: 'Q', 'y': False}),
            matchers.MatchesSetwise(
                matchers.Is(False),
                matchers.Equals(123),
                matchers.Equals('Q')))

    def test_flatten(self):
        self.assertEqual(
            [1, 2, 3, 4],
            self._runner.testFlatten([[1, 2], [3, 4]]))

    def test_dict_get(self):
        self.assertEqual(
            2,
            self._runner.testDictGet({'a': 'x', 'y': 2}, 'y'))

    def test_random_name(self):
        name1 = self._runner.testRandomName()
        name2 = self._runner.testRandomName()

        self.assertIsInstance(name1, str)
        self.assertIsInstance(name2, str)
        self.assertThat(len(name1), matchers.GreaterThan(12))
        self.assertThat(len(name2), matchers.GreaterThan(12))
        self.assertThat(name1, matchers.NotEquals(name2))

    def test_pselect(self):
        self.assertEqual(
            [1, 4, 9, 16, 25, 36, 49, 64, 81, 100],
            self._runner.testPSelect([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]))

    def test_bind(self):
        self.assertEqual(
            {
                'a': 5,
                'b': True,
                'c': {
                    'd': 'x'
                },
                'e': ['qq-123']
            },
            self._runner.testBind(
                {
                    'a': '$a',
                    '$Z': True,
                    'c': {
                        'd': '$value of d'
                    },
                    'e': ['$qq-{suffix}']
                },
                {'a': 5, 'Z': 'b', 'value of d': 'x', 'suffix': 123}))

    def test_patch(self):
        self.assertEqual(
            {'foo': 'bar', 'baz': [42]},
            self._runner.testPatch())

    def test_skip(self):
        self.assertEqual(
            [3, 4, 5, 6],
            self._runner.testSkip([1, 2, 3, 4, 5, 6], 2)
        )

    def test_take(self):
        self.assertEqual(
            [1, 2, 3],
            self._runner.testTake([1, 2, 3, 4, 5, 6], 3)
        )

    def test_skip_take(self):
        self.assertEqual(
            [3, 4, 5],
            self._runner.testSkipTake([1, 2, 3, 4, 5, 6, 7, 8],
                                      2, 3)
        )

    def test_skip_take_chained(self):
        self.assertEqual(
            [3, 4, 5],
            self._runner.testSkipTakeChained(
                [1, 2, 3, 4, 5, 6, 7, 8],
                2, 3)
        )

    def test_aggregate(self):
        self.assertEqual(
            10,
            self._runner.testAggregate([1, 2, 3, 4])
        )

    def test_aggregate_with_initializer(self):
        self.assertEqual(
            21,
            self._runner.testAggregateWithInitializer([1, 2, 3, 4], 11)
        )

    def test_id(self):
        obj_id = self._runner.root.object_id
        self.assertEqual(obj_id * 2, self._runner.testId())

    def test_type(self):
        self.assertEqual('TestEngineFunctions' * 2, self._runner.testType())

    def test_is_operator(self):
        self.assertTrue(self._runner.testIsOperator())
        self.assertFalse(self._runner.testNegativeIsOperator())

    def test_new_object_assignment(self):
        self.assertTrue(self._runner.testNewObjectAssignment())

    @mock.patch('murano.engine.system.yaql_functions.key_manager')
    @mock.patch('murano.engine.system.yaql_functions.castellan_utils')
    def test_decrypt_data(self, mock_castellan_utils, mock_key_manager):
        dummy_context = mock.MagicMock()
        mock_castellan_utils.credential_factory.return_value = dummy_context

        encrypted_value = '91f784d0-5ef1-4b6f-9311-9b5a33d828d8'
        decrypted_value = 'secret_password'

        mock_key_manager.API().get.return_value.get_encoded.return_value =\
            decrypted_value
        self.assertEqual(decrypted_value,
                         self._runner.testDecryptData(encrypted_value))
        mock_key_manager.API().get.assert_called_once_with(dummy_context,
                                                           encrypted_value)

    @mock.patch('murano.engine.system.yaql_functions.LOG')
    def test_decrypt_data_not_configured(self, mock_log):
        encrypted_value = '91f784d0-5ef1-4b6f-9311-9b5a33d828d8'
        self.assertRaises(castellan_exception.AuthTypeInvalidError,
                          self._runner.testDecryptData, encrypted_value)
        mock_log.error.assert_called()



# Copyright (c) 2015 Telefonica I+D.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import fixtures

from murano.db.services import environment_templates as env_temp
from murano.tests.unit import base
from murano.tests.unit.db.services import environment_templates as et


class TestTemplateServices(base.MuranoWithDBTestCase,
                           fixtures.TestWithFixtures):

    def setUp(self):
        super(TestTemplateServices, self).setUp()
        self.template_services = env_temp.EnvTemplateServices
        self.uuids = ['template_id', 'template_id2']
        self.mock_uuid = self._stub_uuid(self.uuids)
        self.addCleanup(mock.patch.stopall)

    def test_create_template(self):
        """Check creating a template without services."""
        fixture = self.useFixture(et.EmptyEnvironmentTemplateFixture())
        """Check the creation of a template."""
        body = {
            "name": "my_template"
        }
        template_des = self.template_services.create(body, 'tenant_id')
        self.assertEqual(fixture.environment_template_desc,
                         template_des.description)

    def test_clone_template(self):
        """Check the clonation of a template."""
        body = {
            "name": "my_template"
        }
        template = self.template_services.create(body, 'tenant_id')
        cloned_template = self.template_services.clone(template['id'],
                                                       'id2',
                                                       "my_template2", False)
        self.assertEqual(cloned_template.description['name'], 'my_template2')
        self.assertEqual(cloned_template.description['tenant_id'], 'id2')

    def test_get_empty_template(self):
        """Check obtaining information about a template without services."""
        fixture = self.useFixture(et.EmptyEnvironmentTemplateFixture())
        self.test_create_template()
        template = \
            self.template_services.get_description("template_id")
        self.assertEqual(fixture.environment_template_desc, template)

    def test_get_template_services(self):
        """Check obtaining information about a template with services."""
        fixture = self.useFixture(et.AppEnvTemplateFixture())
        template = self.template_services.create(fixture.env_template_desc,
                                                 'tenant_id')
        self.assertEqual(fixture.env_template_desc, template.description)
        template_des = \
            self.template_services.get_description("template_id")
        self.assertEqual(fixture.env_template_desc, template_des)

    def test_get_template_no_exists(self):
        """Check obtaining information about a non-existent template

        Check obtaining information about a template which
        does not exist.
        """
        self.assertRaises(ValueError,
                          self.template_services.get_description,
                          'no_exists')

    def test_delete_template(self):
        """Check deleting a template."""
        self.test_create_template()
        self.template_services.delete("template_id")

    def _stub_uuid(self, values=None):
        class FakeUUID(object):
            def __init__(self, v):
                self.hex = v

        if values is None:
            values = []
        mock_uuid4 = mock.patch('uuid.uuid4').start()
        mock_uuid4.side_effect = [FakeUUID(v) for v in values]
        return mock_uuid4

    def test_get_application_description(self):
        fixture = self.useFixture(et.AppEnvTemplateFixture())
        self.template_services.create(fixture.env_template_desc,
                                      'tenant_id')
        template_app_des = (self.template_services.
                            get_application_description("template_id"))
        self.assertEqual(template_app_des,
                         fixture.env_template_desc['services'])

        body = {
            "name": "my_template"
        }
        self.template_services.create(body, 'tenant_id')
        template_app_des_2 = (self.template_services.
                              get_application_description("template_id2"))
        self.assertEqual([], template_app_des_2)



# Copyright (c) 2016 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import semantic_version
from unittest import mock

from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestVersioning(test_case.DslTestCase):
    def test_provided_version(self):
        version = '1.11.111'
        sem_version = semantic_version.Spec('==' + version)
        model = om.Object('Empty', class_version=version)
        m = mock.MagicMock(return_value=self._package_loader._package)
        self._package_loader.load_class_package = m
        self.new_runner(model)
        m.assert_called_once_with('Empty', sem_version)

    def test_empty_provided_version(self):
        version = ''
        sem_version = semantic_version.Spec('>=0.0.0', '<1.0.0')
        model = om.Object('Empty', class_version=version)
        m = mock.MagicMock(return_value=self._package_loader._package)
        self._package_loader.load_class_package = m
        self.new_runner(model)
        m.assert_called_once_with('Empty', sem_version)

    def test_several_in_row(self):
        version = '>3.0.0,<=4.1'
        sem_version = semantic_version.Spec('>3.0.0', '<4.2.0')
        model = om.Object('Empty', class_version=version)
        m = mock.MagicMock(return_value=self._package_loader._package)
        self._package_loader.load_class_package = m
        self.new_runner(model)
        m.assert_called_once_with('Empty', sem_version)



# Copyright (c) 2016 AT&T Corp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslo_messaging.rpc import client
from webob import exc

from murano.api.v1 import schemas
import murano.tests.unit.base as test_base
from murano.tests.unit import utils as test_utils


@mock.patch('murano.api.v1.schemas.policy')
@mock.patch('murano.api.v1.schemas.request_statistics.update_error_count')
@mock.patch('murano.api.v1.schemas.request_statistics.update_count')
class TestSchemas(test_base.MuranoTestCase):

    @classmethod
    def setUpClass(cls):
        super(TestSchemas, cls).setUpClass()
        cls.controller = schemas.Controller()

    @mock.patch('murano.api.v1.schemas.rpc')
    def test_get_schema(self, mock_rpc, *args):
        dummy_context = test_utils.dummy_context()
        dummy_context.GET = {
            'classVersion': 'test_class_version',
            'packageName': 'test_package_name'
        }
        mock_request = mock.MagicMock(context=dummy_context)
        mock_rpc.engine().generate_schema.return_value = 'test_schema'

        result = self.controller.get_schema(mock_request, 'test_class')
        self.assertEqual('test_schema', result)

    @mock.patch('murano.api.v1.schemas.rpc')
    def test_get_schema_negative(self, mock_rpc, *args):
        dummy_context = test_utils.dummy_context()
        dummy_context.GET = {
            'classVersion': 'test_class_version',
            'packageName': 'test_package_name'
        }
        mock_request = mock.MagicMock(context=dummy_context)

        # Test exception handling for pre-defined exception types.
        exc_types = ('NoClassFound',
                     'NoPackageForClassFound',
                     'NoPackageFound')
        for exc_type in exc_types:
            dummy_error = client.RemoteError(exc_type=exc_type,
                                             value='dummy_value')
            mock_rpc.engine().generate_schema.side_effect = dummy_error
            with self.assertRaisesRegex(exc.HTTPNotFound,
                                        dummy_error.value):
                self.controller.get_schema(mock_request, 'test_class')

        # Test exception handling for miscellaneous exception type.
        dummy_error = client.RemoteError(exc_type='TestExcType',
                                         value='dummy_value')
        mock_rpc.engine().generate_schema.side_effect = dummy_error
        with self.assertRaisesRegex(client.RemoteError,
                                    dummy_error.value):
            self.controller.get_schema(mock_request, 'test_class')



# Copyright 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import webob

from keystoneauth1 import exceptions

from murano.api.middleware import ext_context
from murano.tests.unit import base

from oslo_serialization import base64


class MiddlewareExtContextTest(base.MuranoTestCase):

    def test_middleware_ext_context_default(self):
        middleware = ext_context.ExternalContextMiddleware(None)
        middleware.get_keystone_token = mock.MagicMock(return_value="token?")
        auth = 'Basic {}'.format(base64.encode_as_text(b'test:test'))
        request_headers = {
            'Authorization': auth,
        }
        request = webob.Request.blank('/environments',
                                      headers=request_headers)
        middleware.process_request(request)
        self.assertEqual(request.headers.get('X-Auth-Token'), "token?")

    def test_middleware_ext_context_except_key_error(self):
        middleware = ext_context.ExternalContextMiddleware(None)
        middleware.get_keystone_token = mock.MagicMock(
            side_effect=KeyError('test key error')
        )
        auth = 'Basic {}'.format(base64.encode_as_text(b'test:test'))
        request_headers = {
            'Authorization': auth,
        }
        request = webob.Request.blank('/environments',
                                      headers=request_headers)
        self.assertRaises(webob.exc.HTTPUnauthorized,
                          middleware.process_request, request)

    def test_middleware_ext_context_except_unauthorized(self):
        middleware = ext_context.ExternalContextMiddleware(None)
        middleware.get_keystone_token = mock.MagicMock(
            side_effect=exceptions.Unauthorized('')
        )
        auth = 'Basic {}'.format(base64.encode_as_text(b'test:test'))
        request_headers = {
            'Authorization': auth,
        }
        request = webob.Request.blank('/environments',
                                      headers=request_headers)
        self.assertRaises(webob.exc.HTTPUnauthorized,
                          middleware.process_request, request)

    @mock.patch('murano.api.middleware.ext_context.ks_session')
    @mock.patch('murano.api.middleware.ext_context.v3')
    @mock.patch('murano.api.middleware.ext_context.CONF')
    def test_get_keystone_token(self, mock_conf, mock_v3, mock_ks_session):
        mock_ks_session.Session().get_token.return_value = 'test_token'
        test_auth_urls = ['test_url', 'test_url/v2.0', 'test_url/v3']

        for url in test_auth_urls:
            mock_conf.cfapi = mock.MagicMock(auth_url=url,
                                             tenant='test_tenant',
                                             user_domain_name='test_udn',
                                             project_domain_name='test_pdn')

            middleware = ext_context.ExternalContextMiddleware(None)
            token = middleware.get_keystone_token('test_user', 'test_password')
            self.assertEqual('test_token', token)

            expected_kwargs = {
                'auth_url': 'test_url/v3',
                'username': 'test_user',
                'password': 'test_password',
                'project_name': mock_conf.cfapi.tenant,
                'user_domain_name': mock_conf.cfapi.user_domain_name,
                'project_domain_name': mock_conf.cfapi.project_domain_name
            }
            mock_v3.Password.assert_called_once_with(**expected_kwargs)
            mock_v3.Password.reset_mock()

    def test_query_endpoints_except_endpoint_not_found(self):
        middleware = ext_context.ExternalContextMiddleware(None)
        if hasattr(middleware, '_murano_endpoint'):
            setattr(middleware, '_murano_endpoint', None)
        mock_auth = mock.MagicMock()
        mock_auth.get_endpoint.side_effect = exceptions.EndpointNotFound

        middleware._query_endpoints(mock_auth, 'test_session')
        mock_auth.get_endpoint.assert_any_call('test_session',
                                               'application-catalog')

        if hasattr(middleware, '_glare_endpoint'):
            setattr(middleware, '_glare_endpoint', None)
        setattr(middleware, '_murano_endpoint', 'test_endpoint')

        middleware._query_endpoints(mock_auth, 'test_session')
        mock_auth.get_endpoint.assert_any_call('test_session',
                                               'artifact')



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_policy import policy

from murano.common.policies import base

package_policies = [
    policy.DocumentedRuleDefault(
        name='get_package',
        check_str=base.RULE_DEFAULT,
        description="""Returns either detailed package information or
information specific to the package's UI or logo. In addition, checks for the
existence of a given package.""",
        operations=[{'path': '/v1/catalog/packages/{package_id}',
                     'method': 'GET'},
                    {'path': '/v1/catalog/packages',
                     'method': 'GET'},
                    {'path': '/v1/catalog/packages/{package_id}/ui',
                     'method': 'GET'},
                    {'path': '/v1/catalog/packages/{package_id}/logo',
                     'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        name='upload_package',
        check_str=base.RULE_DEFAULT,
        description='Upload a package to the application catalog.',
        operations=[{'path': '/v1/catalog/packages',
                     'method': 'POST'}]),
    policy.DocumentedRuleDefault(
        name='modify_package',
        check_str=base.RULE_DEFAULT,
        description='Update package information for a given package.',
        operations=[{'path': '/v1/catalog/packages/{package_id}',
                     'method': 'PATCH'}]),
    policy.DocumentedRuleDefault(
        name='publicize_package',
        check_str=base.RULE_ADMIN_API,
        description="""Publicize a package across all projects. Grants users in
any project the ability to use the package. Enforced only when `is_public`
parameter is set to True in the request body of the `update` or `upload`
package request.""",
        operations=[{'path': '/v1/catalog/packages/{package_id}',
                     'method': 'PATCH'},
                    {'path': '/v1/catalog/packages',
                     'method': 'POST'}]),
    policy.DocumentedRuleDefault(
        name='manage_public_package',
        check_str=base.RULE_DEFAULT,
        description="""Either update, delete or check for the existence of a
public package. Only enforced when the package is public.""",
        operations=[{'path': '/v1/catalog/packages/{package_id}',
                     'method': 'PATCH'},
                    {'path': '/v1/catalog/packages/{package_id}',
                     'method': 'DELETE'},
                    {'path': '/v1/catalog/packages',
                     'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        name='delete_package',
        check_str=base.RULE_DEFAULT,
        description='Delete a given package.',
        operations=[{'path': '/v1/catalog/packages/{package_id}',
                     'method': 'DELETE'}]),
    policy.DocumentedRuleDefault(
        name='download_package',
        check_str=base.RULE_DEFAULT,
        description='Download a package from the application catalog.',
        operations=[{'path': '/v1/catalog/packages/{package_id}/download',
                     'method': 'GET'}])
]


def list_rules():
    return package_policies



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_db import exception
from oslo_utils import timeutils
import sqlalchemy
from sqlalchemy.sql import func

from murano.db import models
from murano.db import session as db_session


UNCLASSIFIED = 0
APPLICATION = 100
OS_INSTANCE = 200


class InstanceStatsServices(object):
    @staticmethod
    def track_instance(instance_id, environment_id, instance_type,
                       type_name, type_title=None, unit_count=None):

        unit = db_session.get_session()
        try:
            with unit.begin():
                env = unit.query(models.Environment).get(environment_id)
                instance = models.Instance()
                instance.instance_id = instance_id
                instance.environment_id = environment_id
                instance.tenant_id = env.tenant_id
                instance.instance_type = instance_type
                instance.created = timeutils.utcnow_ts()
                instance.destroyed = None
                instance.type_name = type_name
                instance.type_title = type_title
                instance.unit_count = unit_count

                unit.add(instance)
        except exception.DBDuplicateEntry:
            unit.execute(
                sqlalchemy.update(models.Instance).where(
                    models.Instance.instance_id == instance_id and
                    models.Instance.environment_id == environment_id).values(
                        unit_count=unit_count))

    @staticmethod
    def destroy_instance(instance_id, environment_id):
        unit = db_session.get_session()
        instance = unit.query(models.Instance).get(
            (environment_id, instance_id))
        if instance and not instance.destroyed:
            instance.destroyed = timeutils.utcnow_ts()
            instance.save(unit)

    @staticmethod
    def get_aggregated_stats(environment_id):
        unit = db_session.get_session()
        now = timeutils.utcnow_ts()
        query = unit.query(models.Instance.instance_type, func.sum(
            func.coalesce(models.Instance.destroyed, now) -
            models.Instance.created), func.count()).filter(
                models.Instance.environment_id == environment_id)

        res = query.group_by(models.Instance.instance_type).all()

        return [{
                'type': int(record[0]),
                'duration': int(record[1]),
                'count': int(record[2])
                } for record in res]

    @staticmethod
    def get_raw_environment_stats(environment_id, instance_id=None):
        unit = db_session.get_session()
        now = timeutils.utcnow_ts()
        query = unit.query(models.Instance).filter(
            models.Instance.environment_id == environment_id)

        if instance_id:
            query = query.filter(models.Instance.instance_id == instance_id)

        res = query.all()

        return [{
                'type': record.instance_type,
                'duration': (record.destroyed or now) - record.created,
                'type_name': record.type_name,
                'unit_count': record.unit_count,
                'instance_id': record.instance_id,
                'type_title': record.type_title,
                'active': True if not record.destroyed else False
                } for record in res]



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from sqlalchemy import desc
from sqlalchemy.orm import load_only
from webob import exc

from murano.api.v1 import request_statistics
from murano.common.helpers import token_sanitizer
from murano.common import policy
from murano.common import utils
from murano.common import wsgi
from murano.db import models
from murano.db import session as db_session
from murano.utils import check_env

LOG = logging.getLogger(__name__)

API_NAME = 'Deployments'


class Controller(object):
    @request_statistics.stats_count(API_NAME, 'Index')
    def index(self, request, environment_id=None):
        all_environments = environment_id is None
        LOG.debug('Deployments:List <all_environments: {0}>'
                  .format(all_environments))

        if all_environments:
            policy.check("list_deployments_all_environments", request.context)
        else:
            check_env(request, environment_id)
            target = {"environment_id": environment_id}
            policy.check("list_deployments", request.context, target)

        unit = db_session.get_session()

        if all_environments:
            query = unit.query(models.Environment) \
                .options(load_only('tenant_id')) \
                .filter_by(tenant_id=request.context.tenant) \
                .join(models.Task) \
                .order_by(desc(models.Task.created))
            result = query.all()
            # The join statement above fetches the deployments into
            # Environment.tasks. Iterate over that to get the deployments.
            deployments = []
            for row in result:
                for deployment in row.tasks:
                    deployment = set_dep_state(deployment, unit).to_dict()
                    deployments.append(deployment)
        else:
            query = unit.query(models.Task) \
                .filter_by(environment_id=environment_id) \
                .order_by(desc(models.Task.created))
            result = query.all()
            deployments = [set_dep_state(deployment, unit).to_dict()
                           for deployment in result]
        return {'deployments': deployments}

    @request_statistics.stats_count(API_NAME, 'Statuses')
    def statuses(self, request, environment_id, deployment_id):
        target = {"environment_id": environment_id,
                  "deployment_id": deployment_id}
        policy.check("statuses_deployments", request.context, target)

        unit = db_session.get_session()
        query = unit.query(models.Status) \
            .filter_by(task_id=deployment_id) \
            .order_by(models.Status.created)
        deployment = verify_and_get_deployment(unit, environment_id,
                                               deployment_id)

        if 'service_id' in request.GET:
            service_id_set = set(request.GET.getall('service_id'))
            environment = deployment.description
            entity_ids = []
            for service in environment.get('services', []):
                if service['?']['id'] in service_id_set:
                    id_map = utils.build_entity_map(service)
                    entity_ids = entity_ids + id_map.keys()
            if entity_ids:
                query = query.filter(models.Status.entity_id.in_(entity_ids))
            else:
                return {'reports': []}

        result = query.all()
        return {'reports': [status.to_dict() for status in result]}


def _patch_description(description):
    if not description:
        description = {}
    description['services'] = description.pop('applications', [])
    return token_sanitizer.TokenSanitizer().sanitize(description)


def verify_and_get_deployment(db_session, environment_id, deployment_id):
    deployment = db_session.query(models.Task).get(deployment_id)
    if not deployment:
        LOG.error('Deployment with id {id} not found'
                  .format(id=deployment_id))
        raise exc.HTTPNotFound
    if deployment.environment_id != environment_id:
        LOG.error('Deployment with id {d_id} not found in environment '
                  '{env_id}'.format(d_id=deployment_id,
                                    env_id=environment_id))
        raise exc.HTTPBadRequest

    deployment.description = _patch_description(deployment.description)
    return deployment


def create_resource():
    return wsgi.Resource(Controller())


def set_dep_state(deployment, unit):
    num_errors = unit.query(models.Status).filter_by(
        level='error',
        task_id=deployment.id).count()

    num_warnings = unit.query(models.Status).filter_by(
        level='warning',
        task_id=deployment.id).count()

    if deployment.finished:
        if num_errors:
            deployment.state = 'completed_w_errors'
        elif num_warnings:
            deployment.state = 'completed_w_warnings'
        else:
            deployment.state = 'success'
    else:
        if num_errors:
            deployment.state = 'running_w_errors'
        elif num_warnings:
            deployment.state = 'running_w_warnings'
        else:
            deployment.state = 'running'

    deployment.description = _patch_description(deployment.description)
    return deployment



# Copyright (c) 2015 Telefonica I+D.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from webob import exc

from murano.db.services import core_services
from murano.tests.unit import base
from murano.tests.unit.db.services import environment_templates as et


class TestCoreServices(base.MuranoTestCase):
    def setUp(self):
        super(TestCoreServices, self).setUp()
        self.core_services = core_services.CoreServices
        self.addCleanup(mock.patch.stopall)

    @mock.patch('murano.db.services.environment_templates.EnvTemplateServices')
    def test_empty_template(self, template_services_mock):
        """Check obtaining the template description without services."""
        fixture = self.useFixture(et.EmptyEnvironmentTemplateFixture())
        template_services_mock.get_description.return_value = \
            fixture.environment_template_desc
        template_des = self.core_services.get_template_data('any', '/services')
        self.assertEqual([], template_des)
        template_services_mock.get_description.assert_called_with('any')

    @mock.patch('murano.db.services.environment_templates.EnvTemplateServices')
    def test_template_services(self, template_services_mock):
        """Check obtaining the template description with services."""
        fixture_apps = self.useFixture(et.ApplicationsFixture())
        fixture_env_apps = self.useFixture(et.AppEnvTemplateFixture())
        template_services_mock.get_description.return_value = \
            fixture_env_apps.env_template_desc
        template_des = self.core_services.get_template_data('any', '/services')
        self.assertEqual(fixture_apps.applications_desc, template_des)
        template_services_mock.get_description.assert_called_with('any')

    @mock.patch('murano.db.services.environment_templates.EnvTemplateServices')
    def test_template_get_service(self, template_services_mock):
        """Check obtaining the service description."""
        fixture = self.useFixture(et.AppEnvTemplateFixture())
        fixture2 = self.useFixture(et.ApplicationTomcatFixture())
        template_services_mock.get_description.return_value = \
            fixture.env_template_desc
        template_des = \
            self.core_services.get_template_data('any',
                                                 '/services/tomcat_id')
        self.assertEqual(fixture2.application_tomcat_desc, template_des)
        template_services_mock.get_description.assert_called_with('any')

    @mock.patch('murano.db.services.environment_templates.EnvTemplateServices')
    def test_template_post_services(self, template_services_mock):
        """Check adding a service to a template."""
        fixture = self.useFixture(et.EmptyEnvironmentTemplateFixture())
        fixture2 = self.useFixture(et.AppEnvTemplateFixture())
        template_services_mock.get_description.return_value = \
            fixture.environment_template_desc
        template_des = self.core_services.\
            post_env_template_data('any',
                                   fixture2.env_template_desc,
                                   '/services')
        self.assertEqual(fixture2.env_template_desc, template_des)
        template_services_mock.get_description.assert_called_with('any')

    @mock.patch('murano.db.services.environment_templates.EnvTemplateServices')
    def test_template_delete_services(self, template_services_mock):
        """Check deleting a service in a template."""
        fixture2 = self.useFixture(et.AppEnvTemplateFixture())
        fixture = self.useFixture(et.ApplicationTomcatFixture())
        template_services_mock.get_description.return_value = \
            fixture2.env_template_desc
        self.core_services.\
            delete_env_template_data('any',
                                     '/services/54aaa43d-5970')
        template_des = self.core_services.get_template_data('any', '/services')
        self.assertEqual([fixture.application_tomcat_desc], template_des)
        template_services_mock.get_description.assert_called_with('any')

    @mock.patch('murano.db.services.environment_templates.EnvTemplateServices')
    def test_get_template_no_exists(self, template_services_mock):
        """Check obtaining a non-existing service."""
        fixture2 = self.useFixture(et.AppEnvTemplateFixture())

        template_services_mock.get_description.return_value = \
            fixture2.env_template_desc
        self.assertRaises(exc.HTTPNotFound,
                          self.core_services.get_template_data,
                          'any', '/services/noexists')
        template_services_mock.get_description.assert_called_with('any')

    @mock.patch('murano.db.services.environment_templates.EnvTemplateServices')
    def test_adding_services(self, template_services_mock):
        """Check adding services to a template."""
        ftomcat = self.useFixture(et.ApplicationTomcatFixture())
        fmysql = self.useFixture(et.ApplicationMysqlFixture())
        fixture = self.useFixture(et.EmptyEnvironmentTemplateFixture())
        fservices = self.useFixture(et.ApplicationsFixture())
        template_services_mock.get_description.return_value =\
            fixture.environment_template_desc
        self.core_services.\
            post_env_template_data('any',
                                   ftomcat.application_tomcat_desc,
                                   '/services')
        self.core_services.\
            post_env_template_data('any',
                                   fmysql.application_mysql_desc,
                                   '/services')
        template_des = \
            self.core_services.get_template_data('any', '/services')
        self.assertEqual(fservices.applications_desc, template_des)
        template_services_mock.get_description.assert_called_with('any')

    @mock.patch('murano.db.services.environment_templates.EnvTemplateServices')
    def test_none_temp_description(self, template_services_mock):
        fixture = self.useFixture(et.EmptyEnvironmentTemplateFixture())
        template_services_mock.get_description.return_value = None
        self.assertIsNone(self.core_services.
                          get_template_data('any', '/services'))
        self.assertRaises(exc.HTTPNotFound,
                          self.core_services.post_env_template_data,
                          'any', fixture.environment_template_desc,
                          '/services')
        self.assertRaises(exc.HTTPNotFound,
                          self.core_services.post_application_data,
                          'any', fixture.environment_template_desc,
                          '/services')
        self.assertRaises(exc.HTTPNotFound,
                          self.core_services.delete_env_template_data,
                          'any', '/services')
        self.assertRaises(exc.HTTPNotFound,
                          self.core_services.put_application_data,
                          'any', fixture.environment_template_desc,
                          '/services')

    @mock.patch('murano.common.utils.TraverseHelper')
    @mock.patch('murano.db.services.environments.EnvironmentServices')
    def test_post_data(self, env_services_mock, source_mock):
        fixture = self.useFixture(et.EmptyEnvironmentFixture())
        env_services_mock.get_description.return_value = fixture.env_desc
        session_id = None
        ftomcat = self.useFixture(et.ApplicationTomcatFixture())
        self.assertEqual(self.core_services.post_data('any', session_id,
                         ftomcat.application_tomcat_desc,
                         '/services'), ftomcat.application_tomcat_desc)
        self.assertTrue(source_mock.insert.called)

        self.assertEqual(self.core_services.put_data('any', session_id,
                         ftomcat.application_tomcat_desc,
                         '/services'), ftomcat.application_tomcat_desc)
        self.assertTrue(source_mock.update.called)

        self.core_services.delete_data('any', session_id, '/services')
        self.assertTrue(source_mock.remove.called)

    @mock.patch('murano.db.services.environments.EnvironmentServices')
    def test_get_service_status(self, env_services_mock):
        service_id = 12
        fixture = self.useFixture(et.EmptyEnvironmentFixture())
        env_services_mock.get_description.return_value = fixture.env_desc
        self.core_services.get_service_status('any', service_id)
        self.assertTrue(env_services_mock.get_status.called)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import imghdr
import os

import murano.packages.load_utils as load_utils
import murano.tests.unit.base as test_base


class TestHotV1(test_base.MuranoTestCase):

    def test_supplier_info_load(self):
        package_dir = os.path.abspath(
            os.path.join(__file__, '../../test_packages/test.hot.v1.app')
        )
        package = load_utils.load_from_dir(package_dir)

        self.assertIsNotNone(package.supplier)
        self.assertEqual('Supplier Name', package.supplier['Name'])
        self.assertEqual({'Link': 'http://example.com',
                          'Text': 'Example Company'},
                         package.supplier['CompanyUrl'])
        self.assertEqual(
            'Company summary goes here',
            package.supplier['Summary']
        )
        self.assertEqual(
            'Marked up company description goes here',
            package.supplier['Description']
        )
        self.assertEqual('test_supplier_logo.png', package.supplier['Logo'])

        self.assertEqual('png', imghdr.what('', package.supplier_logo))



# Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestAssignments(test_case.DslTestCase):
    def setUp(self):
        super(TestAssignments, self).setUp()
        self._runner = self.new_runner(
            om.Object(
                'SampleClass1',
                stringProperty='string',
                classProperty=om.Object(
                    'SampleClass2',
                    class2Property='another string')))

    def test_assignment(self):
        self.assertEqual(
            {
                'Arr': [5, 2, [10, 123]],
                'Dict': {
                    'Key1': 'V1',
                    'Key2': {'KEY2': 'V3', 'a_b': 'V2'}
                }
            }, self._runner.testAssignment())

    def test_assignment_on_property(self):
        self.assertEqual(
            {
                'Arr': [5, 2, [10, 123]],
                'Dict': {
                    'Key1': 'V1',
                    'Key2': {'KEY2': 'V3', 'a_b': 'V2'}
                }
            }, self._runner.testAssignmentOnProperty())

    def test_assign_by_copy(self):
        self.assertEqual(
            [1, 2, 3],
            self._runner.testAssignByCopy([1, 2, 3]))



# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

import murano.policy.modify.actions.action_manager as am
from murano.policy.modify.actions import default_actions as da
import murano.tests.unit.policy.modify.actions.test_default_actions as tda


class TestActionManager(tda.ModifyActionTestCase):

    def test_loading(self):
        manager = am.ModifyActionManager()
        self.assertEqual(da.RemoveObjectAction,
                         manager.load_action('remove-object'))
        self.assertEqual(da.SetPropertyAction,
                         manager.load_action('set-property'))
        self.assertEqual(da.AddObjectAction,
                         manager.load_action('add-object'))
        self.assertEqual(da.RemoveRelationAction,
                         manager.load_action('remove-relation'))

    def test_caching(self):
        manager = am.ModifyActionManager()
        manager._load_action = mock.MagicMock(wraps=manager._load_action)
        manager.load_action('remove-object')
        # second call is expected to get cached value
        manager.load_action('remove-object')
        manager._load_action.assert_called_once_with('remove-object')

    def test_no_such_action(self):
        manager = am.ModifyActionManager()
        self.assertRaises(ValueError, manager.load_action, 'no-such-action')

    def test_action_apply(self):
        with self._runner.session():
            manager = am.ModifyActionManager()
            obj_id = self._dict_member.id
            action_spec = 'remove-object: {object_id: %s}' % obj_id
            manager.apply_action(self._obj,
                                 action_spec)

    def test_action_apply_invalid_spec(self):
        manager = am.ModifyActionManager()
        self.assertRaises(ValueError,
                          manager.apply_action, self._obj, 'remove-object')



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json

import eventlet
import heatclient.client as hclient
import heatclient.exc as heat_exc
from oslo_config import cfg
from oslo_log import log as logging

from murano.common import auth_utils
from murano.common.helpers import token_sanitizer
from murano.dsl import dsl
from murano.dsl import helpers
from murano.dsl import session_local_storage

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

HEAT_TEMPLATE_VERSION = '2013-05-23'


class HeatStackError(Exception):
    pass


@dsl.name('io.murano.system.HeatStack')
class HeatStack(object):
    def __init__(self, name, description=None, region_name=None):
        self._name = name
        self._template = None
        self._parameters = {}
        self._files = {}
        self._hot_environment = {}
        self._applied = True
        self._description = description
        self._last_stack_timestamps = (None, None)
        self._tags = ''
        self._region_name = region_name
        self._push_thread = None

    def _is_push_thread_alive(self):
        return self._push_thread is not None and not self._push_thread.dead

    def _kill_push_thread(self):
        if self._is_push_thread_alive():
            self._push_thread.cancel()
            self._wait_push_thread()

    def _wait_push_thread(self):
        if not self._is_push_thread_alive():
            return
        self._push_thread.wait()

    @staticmethod
    def _create_client(session, region_name):
        parameters = auth_utils.get_session_client_parameters(
            service_type='orchestration', region=region_name,
            conf='heat', session=session)
        return hclient.Client('1', **parameters)

    @property
    def _client(self):
        return self._get_client(self._region_name)

    @staticmethod
    @session_local_storage.execution_session_memoize
    def _get_client(region_name):
        session = auth_utils.get_client_session(conf='heat')
        return HeatStack._create_client(session, region_name)

    def _get_token_client(self):
        ks_session = auth_utils.get_token_client_session(conf='heat')
        return self._create_client(ks_session, self._region_name)

    def current(self):
        if self._template is not None:
            return self._template
        try:
            stack_info = self._client.stacks.get(stack_id=self._name)
            template = self._client.stacks.template(
                stack_id='{0}/{1}'.format(
                    stack_info.stack_name,
                    stack_info.id))
            self._template = template
            self._parameters.update(
                HeatStack._remove_system_params(stack_info.parameters))
            self._applied = True
            return self._template.copy()
        except heat_exc.HTTPNotFound:
            self._applied = True
            self._template = {}
            self._parameters.clear()
            return {}

    def parameters(self):
        self.current()
        return self._parameters.copy()

    def reload(self):
        self._template = None
        self._parameters.clear()
        return self.current()

    def set_template(self, template):
        self._template = template
        self._parameters.clear()
        self._applied = False

    def set_parameters(self, parameters):
        self._parameters = parameters
        self._applied = False

    def set_files(self, files):
        self._files = files
        self._applied = False

    def set_hot_environment(self, hot_environment):
        self._hot_environment = hot_environment
        self._applied = False

    def update_template(self, template):
        template_version = template.get('heat_template_version',
                                        HEAT_TEMPLATE_VERSION)
        if template_version != HEAT_TEMPLATE_VERSION:
            err_msg = ("Currently only heat_template_version %s "
                       "is supported." % HEAT_TEMPLATE_VERSION)
            raise HeatStackError(err_msg)
        current = self.current()
        self._template = helpers.merge_dicts(self._template, template)
        self._applied = self._template == current and self._applied

    @staticmethod
    def _remove_system_params(parameters):
        return dict((k, v) for k, v in parameters.items() if
                    not k.startswith('OS::'))

    def _get_status(self):
        status = [None]

        def status_func(state_value):
            status[0] = state_value
            return True
        self._wait_state(status_func)
        return status[0]

    def _wait_state(self, status_func, wait_progress=False):
        tries = 4
        delay = 1
        while tries > 0:
            while True:
                try:
                    stack_info = self._client.stacks.get(
                        stack_id=self._name)
                    status = stack_info.stack_status
                    tries = 4
                    delay = 1
                except heat_exc.HTTPNotFound:
                    stack_info = None
                    status = 'NOT_FOUND'
                except Exception:
                    tries -= 1
                    delay *= 2
                    if not tries:
                        raise
                    eventlet.sleep(delay)
                    break

                if 'IN_PROGRESS' in status:
                    eventlet.sleep(2)
                    continue

                last_stack_timestamps = self._last_stack_timestamps
                self._last_stack_timestamps = (None, None) if not stack_info \
                    else(stack_info.creation_time, stack_info.updated_time)

                if (wait_progress and last_stack_timestamps ==
                        self._last_stack_timestamps and
                        last_stack_timestamps != (None, None)):
                    eventlet.sleep(2)
                    continue

                if not status_func(status):
                    reason = ': {0}'.format(
                        stack_info.stack_status_reason) if stack_info else ''
                    raise EnvironmentError(
                        "Unexpected stack state {0}{1}".format(status, reason))

                try:
                    return dict([(t['output_key'], t['output_value'])
                                 for t in stack_info.outputs])
                except Exception:
                    return {}
        return {}

    def output(self):
        self._wait_push_thread()
        return self._wait_state(lambda status: True)

    def _push(self, object_store=None):
        template = copy.deepcopy(self._template)
        s_template = token_sanitizer.TokenSanitizer().sanitize(template)
        LOG.debug('Pushing: {template}'.format(
            template=json.dumps(s_template)))
        object_store = object_store or helpers.get_object_store()
        while True:
            try:
                with helpers.with_object_store(object_store):
                    current_status = self._get_status()
                    resources = template.get('Resources') or template.get(
                        'resources')
                    if current_status == 'NOT_FOUND':
                        if resources is not None:
                            token_client = self._get_token_client()
                            token_client.stacks.create(
                                stack_name=self._name,
                                parameters=self._parameters,
                                template=template,
                                files=self._files,
                                environment=self._hot_environment,
                                disable_rollback=True,
                                tags=self._tags)

                            self._wait_state(
                                lambda status: status == 'CREATE_COMPLETE')
                    else:
                        if resources is not None:
                            self._client.stacks.update(
                                stack_id=self._name,
                                parameters=self._parameters,
                                files=self._files,
                                environment=self._hot_environment,
                                template=template,
                                disable_rollback=True,
                                tags=self._tags)
                            self._wait_state(
                                lambda status: status == 'UPDATE_COMPLETE',
                                True)
                        else:
                            self.delete()
            except heat_exc.HTTPConflict as e:
                LOG.warning('Conflicting operation: {msg}'.format(msg=e))
                eventlet.sleep(3)
            else:
                break

        self._applied = self._template == template

    def push(self, is_async=False):
        if self._applied or self._template is None:
            return

        self._tags = ','.join(CONF.heat.stack_tags)
        if 'heat_template_version' not in self._template:
            self._template['heat_template_version'] = HEAT_TEMPLATE_VERSION

        if 'description' not in self._template and self._description:
            self._template['description'] = self._description

        self._kill_push_thread()
        if is_async:
            if self._push_thread is None:

                def cleanup():
                    try:
                        self._wait_push_thread()
                    finally:
                        self._push_thread = None

                dsl.get_execution_session().on_session_finish(cleanup)

            self._push_thread =\
                eventlet.greenthread.spawn_after(
                    1, self._push, helpers.get_object_store())
        else:
            self._push()

    def delete(self):
        self._kill_push_thread()
        while True:
            try:
                if not self.current():
                    return
                self._wait_state(lambda s: True)
                self._client.stacks.delete(stack_id=self._name)
                self._wait_state(
                    lambda status: status in ('DELETE_COMPLETE', 'NOT_FOUND'),
                    wait_progress=True)
            except heat_exc.NotFound:
                LOG.warning('Stack {stack_name} already deleted?'
                            .format(stack_name=self._name))
                break
            except heat_exc.HTTPConflict as e:
                LOG.warning('Conflicting operation: {msg}'.format(msg=e))
                eventlet.sleep(3)
            else:
                break

        self._template = {}
        self._applied = True



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
import oslo_messaging as messaging
from oslo_messaging.rpc import dispatcher
from oslo_messaging import target

CONF = cfg.CONF

NOTIFICATION_TRANSPORT = None
TRANSPORT = None


def init():
    global TRANSPORT, NOTIFICATION_TRANSPORT
    TRANSPORT = messaging.get_rpc_transport(CONF)
    NOTIFICATION_TRANSPORT = messaging.get_notification_transport(CONF)


def initialized():
    return None not in [TRANSPORT, NOTIFICATION_TRANSPORT]


def cleanup():
    global TRANSPORT, NOTIFICATION_TRANSPORT
    if TRANSPORT is not None:
        TRANSPORT.cleanup()
    if NOTIFICATION_TRANSPORT is not None:
        NOTIFICATION_TRANSPORT.cleanup()
    TRANSPORT = NOTIFICATION_TRANSPORT = None


def get_client(target, timeout=None):
    if TRANSPORT is None:
        init()
    return messaging.RPCClient(
        TRANSPORT,
        target,
        timeout=timeout
    )


def get_server(target, endpoints, executor):
    if TRANSPORT is None:
        init()
    access_policy = dispatcher.DefaultRPCAccessPolicy
    return messaging.get_rpc_server(
        TRANSPORT,
        target,
        endpoints,
        executor=executor,
        access_policy=access_policy
    )


def get_notification_listener(targets, endpoints, executor):
    if NOTIFICATION_TRANSPORT is None:
        init()
    return messaging.get_notification_listener(
        NOTIFICATION_TRANSPORT,
        targets,
        endpoints,
        executor=executor
    )


class ApiClient(object):
    def __init__(self):
        client_target = target.Target('murano', 'results')
        self._client = get_client(client_target, timeout=15)

    def process_result(self, result, environment_id):
        return self._client.call({}, 'process_result', result=result,
                                 environment_id=environment_id)


class EngineClient(object):
    def __init__(self):
        client_target = target.Target('murano', 'tasks')
        self._client = get_client(client_target, timeout=15)

    def handle_task(self, task):
        return self._client.cast({}, 'handle_task', task=task)

    def call_static_action(self, task):
        return self._client.call({}, 'call_static_action', task=task)

    def generate_schema(self, credentials, class_name, method_names=None,
                        class_version=None, package_name=None):
        return self._client.call(
            credentials, 'generate_schema',
            class_name=class_name,
            method_names=method_names,
            class_version=class_version,
            package_name=package_name
        )


def api():
    if not initialized():
        init()
    return ApiClient()


def engine():
    if not initialized():
        init()
    return EngineClient()



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import sys
import weakref

from oslo_log import log as logging
from yaql.language import specs

from murano.common import utils
from murano.dsl import constants
from murano.dsl import dsl
from murano.dsl import dsl_types
from murano.dsl import exceptions
from murano.dsl import helpers
from murano.dsl import macros
from murano.dsl import meta
from murano.dsl import typespec
from murano.dsl import virtual_exceptions
from murano.dsl import yaql_integration


LOG = logging.getLogger(__name__)
macros.register()
virtual_exceptions.register()


class MuranoMethod(dsl_types.MuranoMethod, meta.MetaProvider):
    def __init__(self, declaring_type, name, payload, original_name=None,
                 ephemeral=False):
        self._name = name
        original_name = original_name or name
        self._declaring_type = weakref.ref(declaring_type)
        self._meta_values = None
        self_ref = self if ephemeral else weakref.proxy(self)

        if callable(payload):
            if isinstance(payload, specs.FunctionDefinition):
                self._body = payload
            else:
                self._body = yaql_integration.get_function_definition(
                    payload, self_ref, original_name)
            self._arguments_scheme = None
            self._scope = self._body.meta.get(constants.META_SCOPE)
            if declaring_type.extension_class and any((
                    helpers.inspect_is_static(
                        declaring_type.extension_class, original_name),
                    helpers.inspect_is_classmethod(
                        declaring_type.extension_class, original_name))):
                self._usage = self._body.meta.get(
                    constants.META_USAGE, dsl_types.MethodUsages.Static)
                if self._usage not in dsl_types.MethodUsages.StaticMethods:
                    raise ValueError(
                        'Invalid Usage for static method ' + self.name)
            else:
                self._usage = (self._body.meta.get(constants.META_USAGE) or
                               dsl_types.MethodUsages.Runtime)
                if self._usage not in dsl_types.MethodUsages.InstanceMethods:
                    raise ValueError(
                        'Invalid Usage for instance method ' + self.name)
                self._resolve_usage_and_scope()
            if self._scope is None:
                self._scope = dsl_types.MethodScopes.Session
            if (self._body.name.startswith('#') or
                    self._body.name.startswith('*')):
                raise ValueError(
                    'Import of special yaql functions is forbidden')
            self._meta = meta.MetaData(
                self._body.meta.get(constants.META_MPL_META),
                dsl_types.MetaTargets.Method,
                declaring_type)
        else:
            payload = payload or {}
            self._body = macros.MethodBlock(payload.get('Body'), name)
            self._usage = payload.get(
                'Usage') or dsl_types.MethodUsages.Runtime
            self._scope = payload.get('Scope')
            self._resolve_usage_and_scope()
            if self._scope is None:
                self._scope = dsl_types.MethodScopes.Session
            arguments_scheme = helpers.list_value(payload.get('Arguments'))
            if isinstance(arguments_scheme, dict):
                arguments_scheme = [{key: value} for key, value in
                                    arguments_scheme.items()]
            self._arguments_scheme = collections.OrderedDict()
            seen_varargs = False
            seen_kwargs = False
            args_order_error = False
            for record in arguments_scheme:
                if not isinstance(record, dict) or len(record) > 1:
                    raise exceptions.DslSyntaxError(
                        'Invalid arguments declaration')
                name = list(record.keys())[0]
                argument = MuranoMethodArgument(
                    self, self.name, name, record[name])
                usage = argument.usage
                if (usage == dsl_types.MethodArgumentUsages.Standard and
                        (seen_kwargs or seen_varargs)):
                    args_order_error = True
                elif usage == dsl_types.MethodArgumentUsages.VarArgs:
                    if seen_kwargs or seen_varargs:
                        args_order_error = True
                    seen_varargs = True
                elif usage == dsl_types.MethodArgumentUsages.KwArgs:
                    if seen_kwargs:
                        args_order_error = True
                    seen_kwargs = True

                if args_order_error:
                    raise exceptions.DslSyntaxError(
                        'Invalid argument order in method {0}'.format(
                            self.name))
                else:
                    self._arguments_scheme[name] = argument

            self._meta = meta.MetaData(
                payload.get('Meta'),
                dsl_types.MetaTargets.Method,
                declaring_type)

        self._instance_stub, self._static_stub = \
            yaql_integration.build_stub_function_definitions(self_ref)

    def _resolve_usage_and_scope(self):
        if self._usage == dsl_types.MethodUsages.Action:
            runtime_version = self.declaring_type.package.runtime_version
            if runtime_version > constants.RUNTIME_VERSION_1_3:
                LOG.warning('"Usage: Action" is deprecated, '
                            'use "Scope: Public" instead')
            if self._scope == dsl_types.MethodScopes.Session:
                raise ValueError(
                    'Both "Usage: Action" and "Scope: Session" are '
                    'provided for method ' + self.name)
            self._scope = dsl_types.MethodScopes.Public

    @property
    def name(self):
        return self._name

    @property
    def declaring_type(self):
        return self._declaring_type()

    @property
    def arguments_scheme(self):
        return self._arguments_scheme

    @property
    def instance_stub(self):
        return self._instance_stub

    @property
    def static_stub(self):
        return self._static_stub

    @property
    def usage(self):
        return self._usage

    @usage.setter
    def usage(self, value):
        self._usage = value

    @property
    def scope(self):
        return self._scope

    @scope.setter
    def scope(self, value):
        self._scope = value

    @property
    def body(self):
        return self._body

    @property
    def is_static(self):
        return self.usage in dsl_types.MethodUsages.StaticMethods

    @property
    def is_action(self):
        return (self.scope == dsl_types.MethodScopes.Public or
                self.usage == dsl_types.MethodUsages.Action)

    def get_meta(self, context):
        def meta_producer(cls):
            method = cls.methods.get(self.name)
            if method is None:
                return None
            return method._meta

        if self._meta_values is None:
            executor = helpers.get_executor()
            context = executor.create_type_context(
                self.declaring_type, caller_context=context)
            self._meta_values = meta.merge_providers(
                self.declaring_type, meta_producer, context)
        return self._meta_values

    def __repr__(self):
        return 'MuranoMethod({0}::{1})'.format(
            self.declaring_type.name, self.name)

    def invoke(self, this, args, kwargs, context=None, skip_stub=False):
        if isinstance(this, dsl.MuranoObjectInterface):
            this = this.object
        if this and not self.declaring_type.is_compatible(this):
            raise Exception("'this' must be of compatible type")
        if not this and not self.is_static:
            raise Exception("A class instance is required")

        if isinstance(this, dsl_types.MuranoObject):
            this = this.cast(self.declaring_type)
        else:
            this = self.declaring_type
        executor = helpers.get_executor()
        return executor.invoke_method(
            self, this, context, args, kwargs, skip_stub)


class MuranoMethodArgument(dsl_types.MuranoMethodArgument, typespec.Spec,
                           meta.MetaProvider):
    def __init__(self, murano_method, method_name, arg_name, declaration):
        super(MuranoMethodArgument, self).__init__(
            declaration, murano_method.declaring_type)
        self._method_name = method_name
        self._arg_name = arg_name
        self._murano_method = weakref.ref(murano_method)
        self._meta = meta.MetaData(
            declaration.get('Meta'),
            dsl_types.MetaTargets.Argument, self.murano_method.declaring_type)
        self._usage = declaration.get('Usage') or \
            dsl_types.MethodArgumentUsages.Standard

        if self._usage not in dsl_types.MethodArgumentUsages.All:
            raise exceptions.DslSyntaxError(
                'Unknown usage {0}. Must be one of ({1})'.format(
                    self._usage, ', '.join(dsl_types.MethodArgumentUsages.All)
                ))

    def transform(self, value, this, *args, **kwargs):
        try:
            if self.murano_method.usage == dsl_types.MethodUsages.Extension:
                this = self.murano_method.declaring_type
            return super(MuranoMethodArgument, self).transform(
                value, this, *args, **kwargs)
        except exceptions.ContractViolationException as e:
            msg = u'[{0}::{1}({2}{3})] {4}'.format(
                self.murano_method.declaring_type.name,
                self.murano_method.name, self.name,
                e.path, str(e))
            utils.reraise(exceptions.ContractViolationException,
                          exceptions.ContractViolationException(msg),
                          sys.exc_info()[2])

    @property
    def murano_method(self):
        return self._murano_method()

    @property
    def name(self):
        return self._arg_name

    @property
    def usage(self):
        return self._usage

    def get_meta(self, context):
        executor = helpers.get_executor()
        context = executor.create_type_context(
            self.murano_method.declaring_type, caller_context=context)

        return self._meta.get_meta(context)

    def __repr__(self):
        return 'MuranoMethodArgument({method}::{name})'.format(
            method=self.murano_method.name, name=self.name)



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

from yaql.language import specs
from yaql.language import utils
from yaql.language import yaqltypes

from murano.dsl import constants
from murano.dsl import dsl
from murano.dsl import dsl_types
from murano.dsl import exceptions
from murano.dsl import helpers
from murano.dsl import yaql_functions
from murano.dsl import yaql_integration


def _prepare_context():
    @specs.parameter('name', yaqltypes.StringConstant())
    def get_context_data(context, name):
        root_context = context['#root_context']

        def set_data(value):
            if not name or name == '$' or name == '$this':
                raise ValueError('Cannot assign to {0}'.format(name))
            ctx = root_context
            while constants.CTX_VARIABLE_SCOPE not in ctx:
                ctx = ctx.parent
            ctx[name] = value

        return _Property(lambda: root_context[name], set_data)

    @specs.parameter('this', _Property)
    @specs.parameter('key', yaqltypes.Keyword())
    def attribution(context, this, key):
        def setter(src_property, value):
            src = src_property.get()
            if isinstance(src, utils.MappingType):
                src_property.set(
                    utils.FrozenDict(
                        itertools.chain(
                            src.items(),
                            ((key, value),))))
            elif isinstance(src, dsl_types.MuranoObject):
                src.set_property(key, value, context['#root_context'])
            elif isinstance(src, (
                    dsl_types.MuranoTypeReference,
                    dsl_types.MuranoType)):
                if isinstance(src, dsl_types.MuranoTypeReference):
                    mc = src.type
                else:
                    mc = src
                helpers.get_executor().set_static_property(
                    mc, key, value, context['#root_context'])
            else:
                raise ValueError(
                    'attribution may only be applied to '
                    'objects and dictionaries')

        def getter(src):
            if isinstance(src, utils.MappingType):
                return src.get(key, {})
            elif isinstance(src, dsl_types.MuranoObject):
                try:
                    return src.get_property(key, context['#root_context'])
                except exceptions.UninitializedPropertyAccessError:
                    return {}
            elif isinstance(src, (
                    dsl_types.MuranoTypeReference,
                    dsl_types.MuranoType)):
                if isinstance(src, dsl_types.MuranoTypeReference):
                    mc = src.type
                else:
                    mc = src
                return helpers.get_executor().get_static_property(
                    mc, key, context['#root_context'])
            else:
                raise ValueError(
                    'attribution may only be applied to '
                    'objects and dictionaries')

        return _Property(
            lambda: getter(this.get()),
            lambda value: setter(this, value))

    @specs.parameter('this', _Property)
    @specs.parameter('index', yaqltypes.Lambda(with_context=True))
    def indexation(context, this, index):
        index = index(context['#root_context'])

        def getter(src):
            if utils.is_sequence(src):
                return src[index]
            else:
                raise ValueError('indexation may only be applied to lists')

        def setter(src_property, value):
            src = src_property.get()
            if utils.is_sequence(src):
                src_property.set(src[:index] + (value,) + src[index + 1:])
            elif isinstance(src, utils.MappingType):
                attribution(context, src_property, index).set(value)

        if isinstance(index, int):
            return _Property(
                lambda: getter(this.get()),
                lambda value: setter(this, value))
        else:
            return attribution(context, this, index)

    def _wrap_type_reference(tr, context):
        return _Property(
            lambda: tr, context['#self']._invalid_target)

    @specs.parameter('prefix', yaqltypes.Keyword())
    @specs.parameter('name', yaqltypes.Keyword())
    @specs.name('#operator_:')
    def ns_resolve(context, prefix, name):
        return _wrap_type_reference(
            yaql_functions.ns_resolve(context, prefix, name), context)

    @specs.parameter('name', yaqltypes.Keyword())
    @specs.name('#unary_operator_:')
    def ns_resolve_unary(context, name):
        return _wrap_type_reference(
            yaql_functions.ns_resolve_unary(context, name), context)

    @specs.parameter('object_', dsl_types.MuranoObject)
    def type_(context, object_):
        return _wrap_type_reference(yaql_functions.type_(object_), context)

    @specs.name('type')
    @specs.parameter('cls', dsl.MuranoTypeParameter())
    def type_from_name(context, cls):
        return _wrap_type_reference(cls, context)

    res_context = yaql_integration.create_empty_context()
    res_context.register_function(get_context_data, '#get_context_data')
    res_context.register_function(attribution, '#operator_.')
    res_context.register_function(indexation, '#indexer')
    res_context.register_function(ns_resolve)
    res_context.register_function(ns_resolve_unary)
    res_context.register_function(type_)
    res_context.register_function(type_from_name)
    return res_context


class _Property(object):
    def __init__(self, getter, setter):
        self._getter = getter
        self._setter = setter

    def get(self):
        return self._getter()

    def set(self, value):
        self._setter(value)


class LhsExpression(object):
    lhs_context = _prepare_context()

    def __init__(self, expression):
        self._expression = expression

    def _invalid_target(self, *args, **kwargs):
        raise exceptions.InvalidLhsTargetError(self._expression)

    def __call__(self, value, context):
        new_context = LhsExpression.lhs_context.create_child_context()
        new_context[''] = context['$']
        new_context['#root_context'] = context
        new_context['#self'] = self
        for name in (constants.CTX_NAMES_SCOPE,):
            new_context[name] = context[name]
        prop = self._expression(context=new_context)
        if not isinstance(prop, _Property):
            self._invalid_target()
        prop.set(value)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""WSGI script for murano-api.

Script for running murano-api under Apache2.
"""


from oslo_config import cfg
import oslo_i18n as i18n
from oslo_log import log as logging

from murano.api.v1 import request_statistics
from murano.common import app_loader
from murano.common import config
from murano.common.i18n import _
from murano.common import policy
from murano.common import server


def init_application():
    i18n.enable_lazy()

    LOG = logging.getLogger('murano.api')

    logging.register_options(cfg.CONF)
    # NOTE(hberaud): Call reset to ensure the ConfigOpts object doesn't
    # already contain registered options if the app is reloaded.
    cfg.CONF.reset()
    cfg.CONF(project='murano')
    logging.setup(cfg.CONF, 'murano')
    config.set_middleware_defaults()
    request_statistics.init_stats()
    policy.init()
    server.get_notification_listener().start()
    server.get_rpc_server().start()

    port = cfg.CONF.bind_port
    host = cfg.CONF.bind_host
    LOG.info(_('Starting Murano REST API on %(host)s:%(port)s'),
             {'host': host, 'port': port})
    return app_loader.load_paste_app('murano')



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from murano.dsl.principal_objects import stack_trace


class MuranoPlException(Exception):
    def __init__(self, names, message, stacktrace, extra=None, cause=None):
        super(MuranoPlException, self).__init__(
            u'[{0}]: {1}'.format(', '.join(names), message))
        if not isinstance(names, list):
            names = [names]
        self._names = names
        self._message = message
        self._stacktrace = stacktrace
        self._extra = extra or {}
        self._cause = cause

    @property
    def names(self):
        return self._names

    @property
    def message(self):
        return self._message

    @property
    def stacktrace(self):
        return self._stacktrace

    @property
    def extra(self):
        return self._extra

    @property
    def cause(self):
        return self._cause

    @staticmethod
    def from_python_exception(exception, context):
        stacktrace = stack_trace.create_stack_trace(context)
        exception_type = type(exception)
        builtins_module = 'builtins'
        module = exception_type.__module__
        if module == builtins_module:
            names = [exception_type.__name__]
        else:
            names = ['{0}.{1}'.format(exception_type.__module__,
                                      exception_type.__name__)]

        result = MuranoPlException(
            names, str(exception), stacktrace)
        _, _, exc_traceback = sys.exc_info()
        result.original_exception = exception
        result.original_traceback = exc_traceback
        return result

    def _format_name(self):
        if not self._names:
            return ''
        elif len(self._names) == 1:
            return self._names[0]
        else:
            return self._names

    def format(self, prefix=''):
        text = ('{3}{0}: {1}\n'
                '{3}Traceback (most recent call last):\n'
                '{2}').format(
            self._format_name(), self.message,
            self.stacktrace().toString(prefix + '  '), prefix)
        if self._cause is not None:
            text += '\n\n{0} Caused by {1}'.format(
                prefix, self._cause.format(prefix + ' ').lstrip())
        return text



#    Copyright (C) 2014 Mirantis Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import sys

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('../../'))
sys.path.insert(0, os.path.abspath('../'))
sys.path.insert(0, os.path.abspath('./'))

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.doctest',
              'sphinx.ext.todo',
              'sphinx.ext.coverage',
              'oslo_config.sphinxconfiggen',
              'oslo_config.sphinxext',
              'oslo_policy.sphinxext',
              'oslo_policy.sphinxpolicygen',
              'sphinx.ext.viewcode',
              'sphinxcontrib.httpdomain',]

if not on_rtd:
    extensions.append('openstackdocstheme')

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# openstackdocstheme options
openstackdocs_repo_name = 'openstack/murano'
openstackdocs_pdf_link = True
openstackdocs_bug_project = 'murano'
openstackdocs_bug_tag = ''

config_generator_config_file = '../../etc/oslo-config-generator/murano.conf'
sample_config_basename = '_static/murano'

policy_generator_config_file = [
    ('../../etc/oslo-policy-generator/murano-policy-generator.conf',
     '_static/murano'),
]

# Set the default Pygments syntax
highlight_language = 'python'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['specification/murano-repository.rst',
                    'specification/murano-api.rst',
                    'murano_pl/builtin_functions.rst',
                    'install/configure_network.rst',
                    'articles/ad-ui.rst',
                    'articles/telnet.rst']

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = False

# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

if not on_rtd:
    #TODO(efedorova): Change local theme to correspond with the theme on rtd
    pass

# The name for this set of Sphinx documents. If None, it defaults to
# "<project> v<release> documentation".
html_title = 'Murano'
html_theme = 'openstackdocs'

# Custom sidebar templates, maps document names to template names.
html_sidebars = {
    'index':    ['sidebarlinks.html', 'localtoc.html', 'searchbox.html', 'sourcelink.html'],
    '**':       ['localtoc.html', 'relations.html',
                 'searchbox.html', 'sourcelink.html']
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# -- Options for LaTeX output -------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    ('index', 'doc-murano.tex', u'Murano Documentation',
     u'OpenStack Foundation', 'manual'),
]
latex_domain_indices = False

latex_elements = {
    'makeindex': '',
    'printindex': '',
    'preamble': r'\setcounter{tocdepth}{3}',
    'maxlistdepth': '10',
}

# Disable usage of xindy https://bugzilla.redhat.com/show_bug.cgi?id=1643664
latex_use_xindy = False

# Disable smartquotes, they don't work in latex
smartquotes_excludes = {'builders': ['latex']}



# Copyright (c) 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from murano.cmd import manage
from murano.db.catalog import api as db_catalog_api
from murano.db import models
from murano.db import session as db_session
from murano.tests.unit import base as test_base

CONF = cfg.CONF


class TestManage(test_base.MuranoWithDBTestCase):

    def setUp(self):
        super(TestManage, self).setUp()

        session = db_session.get_session()
        # Create environment.
        self.test_environment = models.Environment(
            name=b'test_environment', tenant_id=b'test_tenant_id',
            version=1
        )
        # Create categories.
        self.test_categories = [
            models.Category(name=b'test_category_1'),
            models.Category(name=b'test_category_2')
        ]
        # Create tags.
        self.test_tags = [
            models.Tag(name=b'test_tag_1'),
            models.Tag(name=b'test_tag_2')
        ]
        # Add environment, categories and tags to DB.
        with session.begin():
            session.add(self.test_environment)
            session.add_all(self.test_categories)
            session.add_all(self.test_tags)
        # Create package.
        self.test_package = models.Package(
            fully_qualified_name=b'test_fqn', name=b'test_name',
            logo=b'test_logo', supplier_logo=b'test_supplier_logo',
            type=b'test_type', description=b'test_desc', is_public=True,
            archive=b'test_archive', ui_definition=b'test_ui_definition',
            categories=self.test_categories, tags=self.test_tags,
            owner_id=self.test_environment.tenant_id,)
        # Add the package to the DB.
        with session.begin():
            session.add(self.test_package)
        # Create class definitions and assign their FKs to test_package.id.
        self.test_class_definitions = [
            models.Class(name=b'test_class_definition_1',
                         package_id=self.test_package.id),
            models.Class(name=b'test_class_definition_2',
                         package_id=self.test_package.id)
        ]
        # Add the class definitions to the DB and update the FK reference for
        # test_package.class_definitions.
        with session.begin():
            session.add_all(self.test_class_definitions)
            self.test_package.class_definitions = self.test_class_definitions
            session.add(self.test_package)
        # Create mock object that resembles loaded package from
        # load_utils.load_from_dir
        self.mock_loaded_package = mock.MagicMock(
            full_name=self.test_package.fully_qualified_name,
            display_name=self.test_package.name,
            package_type=self.test_package.type,
            author=self.test_package.author,
            supplier=self.test_package.supplier,
            description=self.test_package.description,
            tags=[tag.name for tag in self.test_package.tags],
            classes=[cls.name for cls in self.test_package.class_definitions],
            logo=self.test_package.logo,
            supplier_logo=self.test_package.supplier_logo,
            ui=self.test_package.ui_definition,
            blob=self.test_package.archive)

    @mock.patch('murano.cmd.manage.LOG')
    @mock.patch('murano.cmd.manage.load_utils')
    def test_do_import_package(self, mock_load_utils, mock_log):
        manage.CONF = mock.MagicMock()
        manage.CONF.command = mock.MagicMock(
            directory='test_dir',
            categories=[cat.name for cat in self.test_package.categories],
            update=True)
        mock_load_utils.load_from_dir.return_value = self.mock_loaded_package

        manage.do_import_package()

        # Assert that the function ran to completion.
        self.assertIn("Finished import of package",
                      str(mock_log.info.mock_calls[0]))

        # Check that the package was uploaded to the DB.
        filter_params = {
            'name': self.test_package.name,
            'fully_qualified_name': self.test_package.fully_qualified_name,
            'type': self.test_package.type,
            'description': self.test_package.description
        }
        retrieved_package = None
        session = db_session.get_session()
        with session.begin():
            retrieved_package = session.query(models.Package)\
                .filter_by(**filter_params).first()
        self.assertIsNotNone(retrieved_package)
        self.assertNotEqual(self.test_package.id, retrieved_package.id)

    @mock.patch('murano.cmd.manage.LOG')
    @mock.patch('murano.cmd.manage.load_utils')
    @mock.patch('murano.cmd.manage.db_catalog_api')
    def test_do_import_package_without_update(self, mock_db_catalog_api,
                                              mock_load_utils, mock_log):
        mock_db_catalog_api.package_search.return_value =\
            [self.test_package]
        mock_load_utils.load_from_dir.return_value =\
            mock.MagicMock(full_name='test_full_name')
        manage.CONF = mock.MagicMock()
        manage.CONF.command = mock.MagicMock(
            directory='test_dir',
            categories=[],
            update=False)

        manage.do_import_package()

        mock_log.error.assert_called_once_with(
            "Package '{name}' exists ({pkg_id}). Use --update."
                .format(name='test_full_name', pkg_id=self.test_package.id))

    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_do_list_categories(self, mock_stdout):
        expected_output = ">> Murano package categories:* "\
                          "test_category_1* test_category_2"

        manage.do_list_categories()
        self.assertEqual(expected_output,
                         mock_stdout.getvalue().replace('\n', '')
                         .replace('b\'', '').replace('\'', ''))

    @mock.patch('murano.cmd.manage.db_catalog_api')
    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_do_list_categories_with_no_categories(self, mock_stdout,
                                                   mock_db_catalog_api):
        mock_db_catalog_api.category_get_names.return_value = []
        expected_output = "No categories were found"

        manage.do_list_categories()

        self.assertEqual(
            expected_output, mock_stdout.getvalue().replace('\n', ''))

    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_do_add_category(self, mock_stdout):
        manage.CONF = mock.MagicMock()
        manage.CONF.command.category_name = 'test_category_name'

        expected_output = ">> Successfully added category test_category_name"

        manage.do_add_category()

        self.assertEqual(expected_output,
                         mock_stdout.getvalue().replace('\n', ''))

    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_do_add_category_except_duplicate_error(self, mock_stdout):
        manage.CONF = mock.MagicMock()
        manage.CONF.command.category_name = 'test_category_name'

        expected_output = ">> ERROR: Category \'test_category_name\' already "\
                          "exists"

        db_catalog_api.category_add('test_category_name')
        manage.do_add_category()

        self.assertEqual(expected_output,
                         mock_stdout.getvalue().replace('\n', ''))

    def test_add_command_parsers(self):
        mock_parser = mock.MagicMock()
        mock_subparsers = mock.MagicMock()
        mock_subparsers.add_parser.return_value = mock_parser

        manage.add_command_parsers(mock_subparsers)

        mock_subparsers.add_parser.assert_any_call('import-package')
        mock_subparsers.add_parser.assert_any_call('category-list')
        mock_subparsers.add_parser.assert_any_call('category-add')

        mock_parser.set_defaults.assert_any_call(func=manage.do_import_package)
        mock_parser.set_defaults.assert_any_call(
            func=manage.do_list_categories)
        mock_parser.set_defaults.assert_any_call(func=manage.do_add_category)
        self.assertEqual(4, mock_parser.add_argument.call_count)

    @mock.patch('murano.cmd.manage.CONF')
    def test_main_except_runtime_error(self, mock_conf):
        mock_conf.side_effect = RuntimeError

        with self.assertRaisesRegex(SystemExit, 'ERROR:'):
            manage.main()

    @mock.patch('murano.cmd.manage.CONF')
    def test_main_except_general_exception(self, mock_conf):
        mock_conf.command.func.side_effect = Exception

        expected_err_msg = "murano-manage command failed:"

        with self.assertRaisesRegex(SystemExit, expected_err_msg):
            manage.main()



# Copyright (c) 2017 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path


def secure_join(*parts):
    """Secure version of os.path.join(*parts)

    Joins pathname components and ensures that with each join the result
    is a subdirectory of the previous join
    """
    new = prev = ""
    for part in parts:
        new = os.path.normpath(os.path.join(prev, part))
        if len(new) <= len(prev) or prev != "" and not new.startswith(
                prev + os.path.sep):
            raise ValueError('path {0} is not allowed {1}'.format(
                os.path.join(*parts), parts))
        prev = new
    return new



# Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.dsl import dsl_types
from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestDump(test_case.DslTestCase):
    def setUp(self):
        super(TestDump, self).setUp()
        self._runner = self.new_runner(om.Object('dumptests.TestDump'))

    def test_dump_simple_inline(self):
        source = om.Object('dumptests.DumpTarget1',
                           foo='FOO', bar=[40, 41, 42], baz={'BAZ': 99})
        result = self._runner.testDump(source, 'Inline')
        self.assertIn('id', result)
        res = self._get_body(result)
        self.assertEqual('FOO', res['foo'])
        self.assertEqual([40, 41, 42], res['bar'])
        self.assertEqual({'BAZ': 99}, res['baz'])

    def test_dump_simple_serializable(self):
        source = om.Object('dumptests.DumpTarget1',
                           foo='FOO', bar=[40, 41, 42], baz={'BAZ': 99})
        result = self._runner.testDump(source, 'Serializable')
        self.assertIn('?', result)
        self.assertEqual('dumptests.DumpTarget1/0.0.0@tests',
                         result['?']['type'])

    def test_dump_simple_full_mixed(self):
        source = om.Object('dumptests.DumpTarget1',
                           foo='FOO', bar=[40, 41, 42], baz={'BAZ': 99})

        result = self._runner.testDump(source, 'Mixed')
        self.assertIn('?', result)
        self.assertNotIn('classVersion', result['?'])
        self.assertNotIn('package', result['?'])
        self.assertIsInstance(result['?']['type'], dsl_types.MuranoType)
        self.assertEqual('dumptests.DumpTarget1', result['?']['type'].name)

    def test_nested(self):
        n1 = om.Object('dumptests.DumpTarget1', foo='FOO')
        n2 = om.Object('dumptests.DumpTarget1', foo='BAR')
        n3 = om.Object('dumptests.DumpTarget1', foo='BAZ')
        source = om.Object('dumptests.DumpTarget2',
                           nested=n1, another=n2, ref=n3)
        result = self._runner.testDump(source)
        res = self._get_body(result)
        self.assertIsNotNone(res['ref'])
        self.assertIsNotNone(res['another'])
        self.assertIsNotNone(res['nested'])
        self.assertEqual('FOO', self._get_body(res['nested'])['foo'])
        self.assertEqual('BAR', self._get_body(res['another'])['foo'])
        self.assertEqual('BAZ', self._get_body(res['ref'])['foo'])

    def test_same_ref_dump(self):
        nested = om.Object('dumptests.DumpTarget1', foo='FOO')
        source = om.Object('dumptests.DumpTarget2',
                           nested=nested, another=nested, ref=nested)
        result = self._runner.testDump(source)
        res = self._get_body(result)
        string_keys = [k for k in res.keys()
                       if isinstance(res[k], str)]
        obj_keys = [k for k in res.keys()
                    if isinstance(res[k], dict)]
        self.assertEqual(2, len(string_keys))
        self.assertEqual(1, len(obj_keys))
        obj = self._get_body(res[obj_keys[0]])
        self.assertEqual('FOO', obj['foo'])
        for ref_id in string_keys:
            self.assertEqual(res[obj_keys[0]]['id'], res[ref_id])

    def test_dump_with_meta_attributes(self):
        n1 = om.Object('dumptests.DumpTarget1', foo='FOO')
        n2 = om.Object('dumptests.DumpTarget1', foo='Bar')
        source = om.Object('dumptests.DumpTarget3', a=n1, b=n2)
        result = self._runner.testDump(source)
        res = self._get_body(result)
        self._get_body(res['a'])
        self.assertIsInstance(res['b'], str)

    def test_dump_with_inheritance(self):
        source = om.Object('dumptests.DumpTarget4', foo='FOO', qux='QUX')
        result = self._runner.testDump(source)
        res = self._get_body(result)
        self.assertEqual('FOO', res['foo'])
        self.assertEqual('QUX', res['qux'])

    def test_dump_with_inheritance_upcast_ignored(self):
        source = om.Object('dumptests.DumpTarget4', foo='FOO', qux='QUX')
        result = self._runner.testDumpWithUpcast(source, True, True)
        res = self._get_body(result)
        self.assertEqual('FOO', res['foo'])
        self.assertEqual('QUX', res['qux'])

    def test_dump_with_inheritance_upcast_allowed(self):
        source = om.Object('dumptests.DumpTarget4', foo='FOO', qux='QUX')
        result = self._runner.testDumpWithUpcast(source, True, False)
        res = self._get_body(result)
        self.assertEqual('FOO', res['foo'])
        self.assertNotIn('qux', res)

    def _get_body(self, obj):
        body_key = [k for k in obj.keys()
                    if k not in ('id', 'name', 'metadata')][0]
        self.assertIsInstance(body_key, dsl_types.MuranoType)
        return obj[body_key]
